#    Copyright © 2021 Andrei Puchko
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import sys
import os
import random
import string
import threading


if __name__ == "__main__":

    sys.path.insert(0, ".")
    from q2rad.q2rad import main

    main()

from q2rad import Q2Form
from q2rad.q2raddb import q2cursor, num
from q2gui.q2model import Q2Model
from q2gui import q2app
from q2gui.q2dialogs import q2working
from q2gui.q2app import Q2Actions
from q2gui.q2app import Q2Controls
from q2gui.q2utils import int_

import logging
from logging.handlers import TimedRotatingFileHandler

import gettext

_ = gettext.gettext


def q2choice(records=[], title="Make your choice", column_title="Column"):
    setta = Q2Form(title)
    column = list(records[0].keys())[0]
    setta.add_control(column, column_title, datalen=300)
    setta.no_view_action = 1
    model = Q2Model()
    # model.set_records(
    #     [{"table": x} for x in self.q2_app.db_data.db_schema.get_schema_tables()]
    # )
    model.set_records(records)

    setta.set_model(model)
    setta.heap.selected = None

    def make_choice():
        setta.heap.selected = setta.r.__getattr__(column)
        setta.close()

    setta.add_action(
        _("Select"),
        make_choice,
        hotkey="Enter",
        tag="select",
        eof_disabled=1,
    )
    setta.run()
    return setta.heap.selected


def choice_table():
    return q2choice(
        [{"table": x} for x in q2app.q2_app.db_data.db_schema.get_schema_tables()],
        title="Select table",
        column_title="Table",
    )


def choice_column(table):
    return q2choice(
        [{"col": x} for x in q2app.q2_app.db_data.db_schema.get_schema_columns(table)],
        title="Select column",
        column_title="Column",
    )


def choice_form():
    return q2choice(
        [
            x
            for x in q2cursor(
                """
                select name
                from forms
                order by name
                """,
                q2app.q2_app.db_logic,
            ).records()
        ],
        title="Select form",
        column_title="Form name",
    )


def set_logging(log_folder="log"):
    if not os.path.isdir(log_folder):
        os.mkdir(log_folder)
    handler = TimedRotatingFileHandler(
        f"{log_folder}/q2.log", when="midnight", interval=1, backupCount=5, encoding="utf-8"
    )
    formatter = logging.Formatter("%(asctime)s-%(name)s: %(levelname)s: %(message)s")
    handler.setFormatter(formatter)
    logging.basicConfig(handlers=[handler])


class Q2Tasker:
    def __init__(self, title="Working..."):
        self.rez = {}
        self.threads = {}
        self.title = title

    def _worker(self, name):
        self.rez[name] = self.threads[name]["worker"](*self.threads[name]["args"])

    def add(self, worker, *args, name=""):
        if name == "" or name in self.threads:
            name = "".join(
                random.SystemRandom().choice(string.ascii_uppercase + string.digits) for _ in range(5)
            )
        self.threads[name] = {"worker": worker, "args": args}
        self.threads[name]["thread"] = threading.Thread(target=self._worker, args=(name,))
        self.threads[name]["thread"].start()

    def wait(self):
        def _wait(self=self):
            for name in self.threads:
                self.threads[name]["thread"].join()

        q2working(_wait, self.title)
        return self.rez


class Q2_save_and_run:
    def __init__(self) -> None:
        self.dev_actions = Q2Actions()

    def _add_save_and_run(self: Q2Form, save_only=False):
        self.dev_actions.show_main_button = False
        self.dev_actions.show_actions = False
        self.dev_actions.add_action("Save", worker=self._save, hotkey="F2")
        if save_only is False:
            self.dev_actions.add_action("Save and run", worker=self._save_and_run, hotkey="F4")

        self.add_control(
            "save_and_run_actions", "", actions=self.dev_actions, control="toolbar", nogrid=1, migrate=0
        )

    def _save(self):
        if self.crud_mode == "EDIT":
            self.crud_save(close_form=False)

    def _save_and_run(self):
        if self.crud_mode == "EDIT":
            self._save()
            self.run_action("Run")

    def _save_and_run_disable(self):
        if self.crud_mode != "EDIT":
            self.dev_actions.set_disabled("Save and run")
            self.dev_actions.set_disabled("Save")


class auto_filter:
    def __init__(self, table, mem):
        self.table = table
        self.mem = mem
        self.fico = []
        self.mem.ok_button = True
        self.mem.cancel_button = True
        self.mem.add_ok_cancel_buttons()

        self.auto_filter()

    def auto_filter(self):
        cu = q2cursor(
            f"""
                select *
                from `lines`
                where name  = '{self.table}'
                    and migrate<>''
                    and (label <>'' or gridlabel <> '')
                order by seq
            """,
            self.mem.q2_app.db_logic,
        )
        self.mem.controls.add_control("/f")
        for col in cu.records():
            col = Q2Controls.validate(col)
            self.fico.append(cu.r.column)

            if col["datatype"] in ["date"] or col.get("num"):
                self.mem.controls.add_control("/h", cu.r.label, check=1)
                col["label"] = "from"
                co = col["column"]
                col["column"] = co + "1"
                self.mem.controls.add_control(**col)
                col["label"] = "to"
                col["column"] = co + "2"
                self.mem.controls.add_control(**col)
                self.mem.controls.add_control("/s")
                self.mem.controls.add_control("/")
            else:
                col["label"] = cu.r.label
                col["check"] = 1
                self.mem.controls.add_control(**col)
        self.mem.valid = self.valid

    def valid(self):
        where = []
        for x in self.fico:
            where.append(self.prepare_where(x))
        where_string = " and ".join([x for x in where if x])
        q2app.q2_app.run_form("sales", where=where_string)
        return False

    def prepare_where(self, column=None, control1=None, control2=None):
        mem_widgets = self.mem.widgets().keys()
        if control1 is None:
            if column in mem_widgets:
                control1 = column
            elif column + "1" in mem_widgets:
                control1 = column + "1"
        if control1 not in mem_widgets:
            return ""

        if not self.mem.w.__getattr__(control1).is_checked():
            return ""

        date_control = self.mem.controls.c.__getattr__(control1)["datatype"] == "date"
        num_control = self.mem.controls.c.__getattr__(control1).get("num")
        control1_value = self.mem.s.__getattr__(control1)
        if control2 is None:
            if control1.endswith("1"):
                control2 = control1[:-1] + "2"
                control2_value = self.mem.s.__getattr__(control2)
            else:
                control2_value = None
        if date_control:
            if control1_value == "0000-00-00":
                control1_value = ""
            if control2_value == "0000-00-00":
                control2_value = ""
        elif num_control:
            control1_value = num(control1_value)
            if control2_value:
                control2_value = num(control2_value)

        if (control1_value and control2_value is None) or control1_value == control2_value:
            if date_control or num_control:
                return f"{column} = '{control1_value}'"
            else:
                return f"{column} like '%{control1_value}%'"
        elif (control1_value and not control2_value) or (control1_value > control2_value):
            return f"{column} >= '{control1_value}'"
        elif not control1_value and control2_value:
            return f"{column} <= '{control2_value}'"
        elif control1_value and control2_value:
            return f"{column} >= '{control1_value}' and {column}<='{control2_value}'"
        return ""


from q2rad.q2rad import get_report


def grid_print(mem):
    self = q2app.q2_app
    form = mem
    report = get_report(style=get_report().make_style(font_size=self.q2style.font_size))

    report.data_sets["cursor"] = [{"_n_n_n_": x} for x in range(mem.model.row_count())]

    detail_rows = report.new_rows()
    detail_rows.rows["role"] = "table"
    detail_rows.rows["data_source"] = "cursor"

    header_rows = report.new_rows(
        style=report.make_style(text_align="center", font_weight="bold", vertical_align="middle")
    )

    columns = mem.grid_form.get_controls_list("q2grid")[0].get_columns_settings()

    for pos, x in enumerate(columns):
        columns[pos]["pos"] = pos

    columns = {int(x["data"].split(",")[0]): x for x in columns}
    total_width = 0

    for x in sorted(columns.keys()):
        columns[x]["width"] = int_(columns[x]["data"].split(",")[1])
        columns[x]["cwidth"] = "0"
        total_width += columns[x]["width"]

    page_width = round(total_width / (self.dpi() / 2.54), 2) + 3

    if page_width < 26:
        page_height = 297 / 210 * page_width
    else:
        page_height = 210 / 297 * page_width

    report.add_page(page_width=page_width, page_height=page_height)

    for x in list(columns.keys()):
        columns[x]["cwidth"] = "%s%%" % round(columns[x]["width"] / total_width * 100, 2)

    for x in sorted(columns.keys()):
        report.add_column(width=columns[x]["cwidth"])
        header_rows.set_cell(0, x, "%s" % columns[x]["name"])
        if mem.model.meta[x].get("num") and not mem.model.meta[x].get("relation"):
            format = "N"
        else:
            format = ""
        # elif mem.model.meta[x].get("datatype") == "date":
        #     format = "D"
        # else:
        detail_rows.set_cell(
            0,
            x,
            "{form.grid_data(_n_n_n_, %s, True)}" % columns[x]["pos"],
            style=report.make_style(alignment=mem.model.alignments[x]),
            format=format,
        )

    report.set_data(form, "form")

    detail_rows.set_table_header(header_rows)

    report.add_rows(rows=detail_rows)

    report.run()



from q2rad.version import __version__
from q2gui.q2app import Q2App
from q2gui.q2form import Q2Form


from q2gui.q2app import load_q2engine

load_q2engine(globals(), "PyQt6")


from q2rad.q2utils import grid_print

Q2Form.grid_print = grid_print



#    Copyright © 2021 Andrei Puchko
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

if __name__ == "__main__":
    import sys

    sys.path.insert(0, ".")
    from q2rad.q2rad import main

    main()

from q2gui import q2app
from q2gui.q2form import NEW, COPY
from q2gui.q2model import Q2CursorModel
from q2gui.q2dialogs import q2Mess, q2Wait, q2mess, q2working

from q2db.schema import Q2DbSchema
from q2db.db import Q2Db
from q2db.cursor import Q2Cursor

import gettext
import json
import os

from q2rad import Q2App, Q2Form
from q2rad.q2raddb import q2cursor, insert
from q2rad.q2appmanager import AppManager
from q2rad.q2raddb import open_url


SQL_ENGINES = ["MySQl", "Sqlite", "Postgresql"]

_ = gettext.gettext


class Q2AppSelect(Q2Form):
    def __init__(self, title=""):
        super().__init__(_("Select application"))
        self.selected_application = {}
        self.no_view_action = True
        self.autoload_enabled = True

    def on_init(self):
        q2_app: Q2App = q2app.q2_app
        q2_app.clear_menu()
        q2_app.build_menu()
        q2_app.hide_menubar()
        q2_app.hide_toolbar()
        q2_app.hide_statusbar()
        q2_app.hide_tabbar()

        self.db = Q2Db(database_name="q2apps.sqlite")

        self.define_form()

        data_schema = Q2DbSchema()
        for x in self.get_table_schema():
            data_schema.add(**x)

        self.db.set_schema(data_schema)

    def define_form(self):
        self.add_control("uid", "", datatype="int", pk="*", noform=1, nogrid=1)
        if self.add_control("/h"):
            self.add_control("ordnum", _("Order"), datatype="int")
            self.add_control(
                "autoselect",
                label=_("Autoload"),
                datatype="char",
                datalen=1,
                control="check",
            )
            self.add_control("/")
        self.add_control("name", _("Name"), datatype="char", datalen=100)

        self.add_control("/")
        if self.add_control("/f", _("Data storage")):

            def driverDataValid(self=self):
                self.w.host_data.set_enabled(self.s.driver_data != "Sqlite")
                self.w.port_data.set_enabled(self.s.driver_data != "Sqlite")
                self.w.select_data_storage_file.set_enabled(
                    self.s.driver_data == "Sqlite"
                )

            self.add_control(
                "driver_data",
                label=_("Storage type"),
                gridlabel=_("Data storage type"),
                control="radio",
                datatype="char",
                datalen=30,
                pic=";".join(SQL_ENGINES),
                valid=driverDataValid,
            )
            if self.add_control("/h"):
                self.add_control(
                    "database_data",
                    "Database",
                    gridlabel=_("Data storage"),
                    datatype="char",
                    datalen=100,
                )
                self.add_control(
                    "select_data_storage_file",
                    _("?"),
                    mess=_("Open Data Storage sqlite database file"),
                    control="button",
                    datalen=3,
                    valid=self.openSqliteDataFile,
                )
                self.add_control("/")
            if self.add_control("/h"):
                self.add_control(
                    "host_data", _("Host"), gridlabel=_("Data host"), datalen=100
                )
                self.add_control(
                    "port_data", _("Port"), gridlabel=_("Data port"), datatype="int"
                )
                self.add_control(
                    "guest_mode",
                    _("Guest mode"),
                    control="check",
                    datatype="char",
                    datalen=1,
                    mess=_("No database schema changes"),
                )
                self.add_control("/")

            self.add_control("/")

        if self.add_control("/f", _("Logic storage")):

            def driverLogicValid(form=self):
                form.w.host_logic.set_enabled(form.s.driver_logic != "Sqlite")
                form.w.port_logic.set_enabled(form.s.driver_logic != "Sqlite")
                form.w.select_app_storage_file.set_enabled(
                    form.s.driver_logic == "Sqlite"
                )

            self.add_control(
                "driver_logic",
                label=_("Storage type"),
                gridlabel=_("Logic storage type"),
                control="radio",
                datatype="char",
                datalen=30,
                pic=";".join(SQL_ENGINES),
                valid=driverLogicValid,
            )
            if self.add_control("/h"):

                self.add_control(
                    "database_logic",
                    "Database",
                    gridlabel="Logic storage",
                    datatype="char",
                    datalen=100,
                )
                self.add_control(
                    "select_app_storage_file",
                    _("?"),
                    mess=_("Open App Storage sqlite database file"),
                    control="button",
                    datalen=3,
                    valid=self.openSqliteDataFile,
                )
                self.add_control("/")
            if self.add_control("/h"):
                self.add_control(
                    "host_logic", _("Host"), gridlabel=_("Logic host"), datalen=100
                )
                self.add_control(
                    "port_logic", _("Port"), gridlabel=_("Logic port"), datatype="int"
                )
                self.add_control(
                    "dev_mode",
                    _("Dev mode"),
                    control="check",
                    datatype="char",
                    datalen=1,
                    mess=_("Allow to change App"),
                )
                self.add_control("/")

        self.add_action(
            _("Select"),
            self.select_application,
            hotkey="Enter",
            tag="select",
            eof_disabled=1,
        )

        self.add_action(
            _("Autoload"),
            self.set_autoload,
            icon="💚",
            mess="Toggle autoload mark",
            eof_disabled=1,
            tag="#4dd0e1"
        )

        self.add_action(_("Demo"), self.run_demo)

        self.before_form_show = self.before_form_show
        self.before_crud_save = self.before_crud_save

        cursor: Q2Cursor = self.db.table(table_name="applications", order="ordnum")
        model = Q2CursorModel(cursor)
        self.set_model(model)

        self.actions.add_action("/crud")

    def set_autoload(self):
        clean_this = self.r.autoselect
        self.db.cursor("update applications set autoselect='' ")
        if not clean_this:
            self.db.cursor("update applications set autoselect='*' where uid=%s" % self.r.uid)
        self.refresh(soft=True)

    def openSqliteDataFile(self):
        fname = self.q2_app.get_save_file_dialoq(
            self.focus_widget().meta.get("mess"),
            "",
            _("SQLite (*.sqlite);;All files(*.*)"),
            confirm_overwrite=False,
        )[0]
        if fname:
            if "_app_" in self.focus_widget().meta.get("column"):
                self.s.database_logic = fname
            else:
                self.s.database_data = fname

    def before_grid_show(self):
        self.q2_app.sleep(0.2)
        if self.q2_app.keyboard_modifiers() != "":
            return
        # if self.autoload_enabled:
        #     cu = q2cursor("select * from applications where autoselect<>''", self.db)
        #     if cu.row_count() > 0:
        #         self._select_application(cu.record(0))
        #         return False
        if self.db.table("applications").row_count() <= 0:
            if not os.path.isdir("q2rad_sqlite_databases"):
                os.mkdir("q2rad_sqlite_databases")
            insert(
                "applications",
                {
                    "ordnum": 1,
                    "name": "My first app",
                    "driver_data": "Sqlite",
                    "database_data": "q2rad_sqlite_databases/my_first_app_data_storage.sqlite",
                    "driver_logic": "Sqlite",
                    "database_logic": "q2rad_sqlite_databases/my_first_app_logic_storage.sqlite",
                    "dev_mode": "*"
                },
                self.db,
            )
            self.refresh()

    def before_crud_save(self):
        if self.s.name == "":
            q2Mess(_("Give me some NAME!!!"))
            self.w.name.set_focus()
            return False
        if self.s.database_data == "":
            q2Mess(_("Give me some database!!!"))
            self.w.database_data.set_focus()
            return False
        if self.s.database_logic == "":
            q2Mess(_("Give me some database!!!"))
            self.w.database_logic.set_focus()
            return False
        if self.s.driver_logic == "Sqlite":
            self.s.host_logic = ""
            self.s.port_logic = ""
        if self.s.driver_data == "Sqlite":
            self.s.host_data = ""
            self.s.port_data = ""
        if self.s.autoselect:
            self.db.cursor("update applications set autoselect='' ")
        return True

    def before_form_show(self):
        if self.crud_mode == "NEW":
            self.s.driver_logic = "Sqlite"
            self.s.driver_data = "Sqlite"
            self.s.dev_mode = "*"
        self.w.driver_data.valid()
        self.w.driver_logic.valid()
        if self.crud_mode in [NEW, COPY]:
            self.s.ordnum = self.model.cursor.get_next_sequence("ordnum", self.r.ordnum)
            self.w.name.set_focus()

    def run_demo(self):
        row = {
            "driver_data": "Sqlite",
            "database_data": ":memory:",
            "driver_logic": "Sqlite",
            "database_logic": ":memory:",
            "dev_mode": "",
        }
        self._select_application(row)
        self.q2_app.migrate_db_logic()
        self.q2_app.migrate_db_logic()

        demo_app_url = f"{self.q2_app.q2market_url}/demo_app.json"
        demo_data_url = f"{self.q2_app.q2market_url}/demo_data.json"
        response_app = open_url(demo_app_url)
        response_data = open_url(demo_data_url)
        if response_app and response_data:
            self.close()
            AppManager.import_json_app(json.load(response_app))
            self.q2_app.open_selected_app()
            self.q2_app.migrate_db_data()
            AppManager.import_json_data(json.load(response_data))
        else:
            q2Mess(_("Can't to load Demo App"))

    def _select_application(self, app_data={}):
        q2_app: Q2App = q2app.q2_app
        q2_app.dev_mode = app_data.get("dev_mode")
        q2_app.selected_application = app_data
        q2_app.open_databases()
        q2_app.show_menubar()
        q2_app.show_toolbar()
        q2_app.show_statusbar()
        q2_app.show_tabbar()
        self.q2_app.process_events()

    def select_application(self):
        self.close()
        self.q2_app.process_events()
        self._select_application(self.model.get_record(self.current_row))

    def run(self, autoload_enabled=True):
        self.autoload_enabled = autoload_enabled
        if autoload_enabled:
            cu = q2cursor("select * from applications where autoselect<>''", self.db)
            if cu.row_count() > 0:
                self._select_application(cu.record(0))
                return False
        super().run()



#    Copyright © 2021 Andrei Puchko
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

if __name__ == "__main__":
    import sys

    sys.path.insert(0, ".")
    from q2rad.q2rad import main

    main()

from q2rad import Q2Form
from q2rad.q2raddb import int_
import logging

_logger = logging.getLogger(__name__)


class AppStyleSettings(Q2Form):
    def __init__(self, title=""):
        super().__init__("Style Settings")

    def on_init(self):
        self.color_modes = {"dark": "Dark", "light": "Light", "clean": "Clean"}
        self.last_font_size = self.q2_app.q2style.font_size
        # print("==", self.last_font_size)
        self.last_color_mode = self.q2_app.q2style.color_mode
        self.add_control(
            "color_mode",
            "Color mode",
            pic="System defaulf;Dark;Light;Clean",
            datatype="char",
            control="radio",
            datalen=10,
            valid=self.style_valid,
            data=self.color_modes.get(
                self.q2_app.q2style.color_mode, self.q2_app.q2style.get_system_color_mode()
            ),
        )

        self.add_control("/h", "Font size")
        self.add_control("minus", "-", datatype="int", control="button", valid=self._font_minus)

        self.add_control(
            "font_size",
            "",
            datalen=6,
            datatype="int",
            control="line",
            data=self.q2_app.q2style.font_size,
            valid=self.style_valid,
        )
        self.add_control("plus", "+", datatype="int", control="button", valid=self._font_plus)
        self.add_control("/s")
        self.add_control("/")

        self.ok_button = 1
        self.cancel_button = 1

    def _font_plus(self):
        self.s.font_size = int_(self.s.font_size) + 1
        self.style_valid()

    def _font_minus(self):
        self.s.font_size = int_(self.s.font_size) - 1
        self.style_valid()

    def get_color_mode(self):
        color_mode = self.s.color_mode.lower()
        if color_mode not in ("dark", "light", "clean"):
            color_mode = None
        return color_mode

    def valid(self):
        self.style_valid()
        color_mode = self.get_color_mode()
        self.q2_app.q2style.font_size = int_(self.s.font_size)
        self.q2_app.settings.set("Style Settings", "color_mode", color_mode)
        self.q2_app.settings.set("Style Settings", "font_size", self.s.font_size)
        self.q2_app.set_color_mode(color_mode)

    def style_valid(self):
        self.q2_app.q2style.font_size = int_(self.s.font_size)
        if self.q2_app.q2style.font_size < 8:
            self.s.font_size = 8
            self.style_valid()
        self.q2_app.set_color_mode(self.get_color_mode())
        self.q2_app.set_color_mode(self.get_color_mode())

    def close(self):
        if not self.ok_pressed:
            self.q2_app.q2style.font_size = self.last_font_size
            self.q2_app.set_color_mode(self.last_color_mode)
        return super().close()
